# -*- coding: utf-8 -*-
# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import json
import re
import time

from djankiserv.unki import checksum  # checksum never actually used, so all occurrences dead code
from djankiserv.unki import splitFields

MODEL_CLOZE = 1


class ModelManager:
    def __init__(self, col):
        self.col = col
        self.needs_saving = False  # to make pylint happy
        self.models = None  # to make pylint happy

    def load(self, json_):
        "Load registry from JSON."
        self.needs_saving = False
        self.models = json.loads(json_)

    def flush(self):
        "Flush the registry if any models were changed."
        if self.needs_saving:
            self.col.db.execute(f"update {self.col.username}.col set models = %s", json.dumps(self.models))
            self.needs_saving = False

    def get(self, mid):
        "Get model with ID, or None."
        return self.models.get(str(mid))

    def all(self):
        "Get all models."
        return list(self.models.values())

    def ensure_name_unique(self, model):
        for mcur in self.all():
            if mcur["name"] == model["name"] and mcur["id"] != model["id"]:
                model["name"] += "-" + checksum(str(time.time()))[:5]
                break

    def update(self, model):
        "Add or update an existing model. Used for syncing and merging."
        self.ensure_name_unique(model)
        self.models[str(model["id"])] = model
        # mark registry changed, but don't bump mod time
        self.needs_saving = True

    def ids(self):
        return list(self.models.keys())

    @staticmethod
    def field_names(model):
        return [field["name"] for field in model["flds"]]

    @staticmethod
    def sort_idx(model):
        return model["sortf"]

    def by_name(self, name):
        "Get model with NAME."
        for model in list(self.models.values()):
            if model["name"] == name:
                return model
        return None

    # FIXME: this shouldn't be necessary!!!
    def set_current(self, model):
        self.col.conf["curModel"] = model["id"]
        self.col.db.mod = True

    def avail_ords(self, model, flds):  # noqa: C901
        "Given a joined field string, return available template ordinals."
        if model["type"] == MODEL_CLOZE:
            return self._avail_cloze_ords(model, flds)
        fields = {}
        for c, f in enumerate(splitFields(flds)):
            fields[c] = f.strip()
        avail = []
        for ordi, otype, req in model["req"]:
            # unsatisfiable template
            if otype == "none":
                continue
            # AND requirement?
            if otype == "all":
                ok = True
                for idx in req:
                    if not fields[idx]:
                        # missing and was required
                        ok = False
                        break
                if not ok:
                    continue
            # OR requirement?
            elif otype == "any":
                ok = False
                for idx in req:
                    if fields[idx]:
                        ok = True
                        break
                if not ok:
                    continue
            avail.append(ordi)
        return avail

    def _avail_cloze_ords(self, model, flds, allow_empty=True):
        sflds = splitFields(flds)
        amap = self.field_map(model)
        ords = set()
        matches = re.findall("{{[^}]*?cloze:(?:[^}]?:)*(.+?)}}", model["tmpls"][0]["qfmt"])
        matches += re.findall("<%cloze:(.+?)%>", model["tmpls"][0]["qfmt"])
        for fname in matches:
            if fname not in amap:
                continue
            ordi = amap[fname][0]
            ords.update([int(stg) - 1 for stg in re.findall(r"(?s){{c(\d+)::.+?}}", sflds[ordi])])
        if -1 in ords:
            ords.remove(-1)
        if not ords and allow_empty:
            # empty clozes use first ord
            return [0]
        return list(ords)

    @staticmethod
    def field_map(model):
        "Mapping of field name -> (ord, field)."
        return dict((field["name"], (field["ord"], field)) for field in model["flds"])



# -*- coding: utf-8 -*-

import io
import json
import logging
import os
import time
import zipfile

from django.conf import settings
from django.contrib.auth import authenticate
from django.contrib.sessions.backends.db import SessionStore
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import AllowAny

from djankiserv.sync import SyncCollectionHandler, full_download, full_upload
from djankiserv.unki import get_data
from djankiserv.unki.collection import Collection
from djankiserv.unki.database import dump_io_to_file

logger = logging.getLogger("djankiserv.views")

## In order to be compatible with the clients and use the bespoke auth, we need to tell DRF not to manage
## auth for the xSYNC methods


def get_session(request):
    return SessionStore(session_key=(request.POST.get("k") or request.GET.get("k") or request.POST.get("sk")))


def safe_get_session(request):
    # Get and verify the session
    print_request(request)

    session = get_session(request)

    if not session.get("skey"):  # FIXME: there is probably a better way than this
        raise PermissionDenied

    return session


def get_collection(session):
    return Collection(session["name"], settings.DJANKISERV_DATA_ROOT)


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_meta(request):
    session = safe_get_session(request)

    if settings.DJANKISERV_GENERATE_TEST_ASSETS:
        session["dump_base"] = os.path.join(settings.DJANKISERV_GENERATE_TEST_ASSETS_DIR, str(time.time()))
        session.save()

    with get_collection(session) as col:
        return JsonResponse(
            {
                "cont": True,
                "hostNum": 1,
                "mod": col.mod,
                "msg": "",
                "musn": col.last_media_usn(),
                "scm": col.scm,
                "ts": int(time.time()),
                "uname": session["name"],
                "usn": col.usn,
            }
        )


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_start(request):
    session = safe_get_session(request)
    data = get_data(request)

    dump_io_to_file(session, "start", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col)

        # FIXME: for the moment we are just ignoring the 'offset' parameter, hoping it will just work if
        # we only use the V1 scheduler
        output = col_handler.start(min_usn=data.get("minUsn"), lnewer=data.get("lnewer"), offset=data.get("offset"))

        ## The following gets values that are required for subsequent calls in a sync flow
        ## as a side-effect of the start. This obviously needs to be completely rethought!!!
        ##
        ## We also need the dump_base in the media so we set for a first time in meta, as start isn't
        ## called if there is no main db sync needed
        ##
        ## FIXME: this needs serious testing for at least:
        ## - failure and then start again with a new sync
        session["min_usn"] = col_handler.min_usn
        session["max_usn"] = col_handler.max_usn
        session["lnewer"] = col_handler.lnewer
        session.save()

        resp = JsonResponse(output)

    dump_io_to_file(session, "start", resp)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_applyGraves(request):
    session = safe_get_session(request)
    data = get_data(request)

    dump_io_to_file(session, "applyGraves", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col, session=session)
        col_handler.applyGraves(chunk=data.get("chunk"))

        resp = JsonResponse({})
    dump_io_to_file(session, "applyGraves", resp)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_applyChanges(request):
    session = safe_get_session(request)
    data = get_data(request)

    dump_io_to_file(session, "applyChanges", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col, session=session)

        ## tests for doing it by chunking rather than all in one go
        # output = col_handler.applyChanges(changes=data.get("changes"))
        # session["tablesLeft"] = col_handler.tablesLeft
        # cache.set(session.skey, session)

        output = col_handler.applyChanges(changes=data.get("changes"))
        resp = JsonResponse(output)

    dump_io_to_file(session, "applyChanges", resp)
    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_chunk(request):
    session = safe_get_session(request)  # performs auth that raises an error if not auth'ed

    dump_io_to_file(session, "chunk", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col, session=session)

        ## FIXME: this is where  the chunking needs to happen. The original call was to col_handler.chunk()
        ## which is from a persistent thread which has a database cursor that gets called with chunks of up
        ## to 250 lines, and it can be called as many times as required until there are none left. This is
        ## obviously very stupid in our context. The all_data_to_sync_down updates the db to say that everything
        ## has already been transferred, though in theory the same logic could be used as the cursor but using
        ## a cache value (we can cache the rows returned but not the cursor)

        all_new_data = col_handler.all_data_to_sync_down()

        resp = JsonResponse(all_new_data, safe=False)

    dump_io_to_file(session, "chunk", resp)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_applyChunk(request):
    session = safe_get_session(request)
    data = get_data(request)

    dump_io_to_file(session, "applyChunk", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col, session=session)

        col_handler.applyChunk(chunk=data.get("chunk"))  # is a void

        resp = JsonResponse({})
    dump_io_to_file(session, "applyChunk", resp)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_sanityCheck2(request):
    session = safe_get_session(request)
    data = get_data(request)

    dump_io_to_file(session, "sanityCheck2", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col, session=session)
        output = col_handler.sanityCheck2(client=data.get("client"))
        resp = JsonResponse(output)
    dump_io_to_file(session, "sanityCheck2", resp)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_finish(request):
    session = safe_get_session(request)

    dump_io_to_file(session, "finish", request)

    with get_collection(session) as col:
        col_handler = SyncCollectionHandler(col, session=session)

        resp = HttpResponse(col_handler.finish())
    dump_io_to_file(session, "finish", resp)

    return resp


@csrf_exempt
@api_view(["POST", "GET"])
@permission_classes((AllowAny,))
def base_hostKey(request):
    data = get_data(request)

    print_request(request)

    username = data.get("u")
    password = data.get("p")

    if not authenticate(username=username, password=password):
        raise PermissionDenied

    s = SessionStore()
    s.create()
    s["skey"] = s.session_key
    s["name"] = username

    s.save()

    return JsonResponse({"key": s.session_key, "hostNum": 1, "host_number": 2})


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_upload(request):
    session = safe_get_session(request)

    dump_io_to_file(session, "upload", request)

    db_bytes = get_data(request)["data"]

    resp = HttpResponse(full_upload(db_bytes, session["name"]))
    dump_io_to_file(session, "upload", resp)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def base_download(request):
    session = safe_get_session(request)

    dump_io_to_file(session, "download", request)

    with get_collection(session) as col:
        resp = HttpResponse(full_download(col, session["name"]))
    dump_io_to_file(session, "download", resp)

    return resp


@csrf_exempt
@api_view(["GET", "POST"])
@permission_classes((AllowAny,))
def media_begin(request):
    session = safe_get_session(request)

    if settings.DJANKISERV_GENERATE_TEST_ASSETS:
        session["dump_base_media"] = os.path.join(
            settings.DJANKISERV_GENERATE_TEST_ASSETS_DIR, "media", str(time.time())
        )
        session.save()

    with get_collection(session) as col:
        data = {
            "data": {"sk": session["skey"], "usn": col.last_media_usn()},
            "err": "",
        }
        resp = JsonResponse(data)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def media_mediaChanges(request):
    session = safe_get_session(request)

    dump_io_to_file(session, "mediaChanges", request, is_media=True)

    with get_collection(session) as col:
        data = get_data(request)

        resp = JsonResponse({"data": col.media_changes(data["lastUsn"]), "err": ""})
    dump_io_to_file(session, "mediaChanges", resp, is_media=True)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def media_mediaSanity(request):
    session = safe_get_session(request)

    dump_io_to_file(session, "mediaSanity", request, is_media=True)

    with get_collection(session) as col:
        if col.media_count() == get_data(request)["local"]:
            resp = JsonResponse({"data": "OK", "err": ""})
        else:
            resp = JsonResponse({"data": "FAILED", "err": ""})

    dump_io_to_file(session, "mediaSanity", resp, is_media=True)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def media_uploadChanges(request):
    session = safe_get_session(request)

    dump_io_to_file(session, "uploadChanges", request, is_media=True)

    with get_collection(session) as col:
        data = get_data(request)["data"]

        """
        The zip file contains files the client hasn't synced with the server
        yet ('dirty'), and info on files it has deleted from its own media dir.
        """
        with zipfile.ZipFile(io.BytesIO(data), "r") as z:
            col.check_zip_data(z)
            processed_count = col.adopt_media_changes_from_zip(z)

        resp = JsonResponse({"data": [processed_count, col.last_media_usn()], "err": ""})
    dump_io_to_file(session, "uploadChanges", resp, is_media=True)

    return resp


@csrf_exempt
@api_view(["POST"])
@permission_classes((AllowAny,))
def media_downloadFiles(request):
    SYNC_ZIP_SIZE = int(2.5 * 1024 * 1024)
    SYNC_ZIP_COUNT = 25

    session = safe_get_session(request)
    data = get_data(request)

    dump_io_to_file(session, "downloadFiles", request, is_media=True)

    with get_collection(session) as col:
        files = data["files"]
        flist = {}
        cnt = 0
        sz = 0
        f = io.BytesIO()

        with zipfile.ZipFile(f, "w", compression=zipfile.ZIP_DEFLATED) as z:
            for fname in files:
                z.write(os.path.join(col.media_dir(), fname), str(cnt))
                flist[str(cnt)] = fname
                sz += os.path.getsize(os.path.join(col.media_dir(), fname))
                if sz > SYNC_ZIP_SIZE or cnt > SYNC_ZIP_COUNT:
                    break
                cnt += 1

            z.writestr("_meta", json.dumps(flist))

        resp = HttpResponse(f.getvalue())
    dump_io_to_file(session, "downloadFiles", resp, is_media=True)

    return resp


# DEBUG AND TESTING STUFF
def print_request(request):
    try:
        if settings.DJANKISERV_DEBUG:
            print(pretty_request(request))
    except NameError:
        pass  # we haven't defined it, so can't be interested


def pretty_request(request):

    headers = ""
    for header, value in request.META.items():
        if not header.startswith("HTTP"):
            continue
        # header = "-".join([h.capitalize() for h in header[5:].lower().split("_")])
        header = "-".join([h.capitalize() for h in header.lower().split("_")])
        headers += "{}: {}\n".format(header, value)

    return (
        "{path}\n"
        "{method} HTTP/1.1\n"
        "Content-Length: {content_length}\n"
        "Content-Type: {content_type}\n"
        "{headers}\n\n"
        "{body}\n\n"
    ).format(
        path=request.path,
        method=request.method,
        content_length=request.META.get("CONTENT_LENGTH"),
        content_type=request.META.get("CONTENT_TYPE"),
        headers=headers,
        body=request.data,
    )


# END DEBUG AND TESTING STUFF



# -*- coding: utf-8 -*-
# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import json
import random
import string

import djankiserv.unki

from . import fieldChecksum, intTime, joinFields, splitFields, stripHTMLMedia


def guid64():
    extra = "!#$%&()*+,-./:;<=>?@[]^_`{|}~"
    num = random.randint(0, 2 ** 64 - 1)
    s = string
    table = s.ascii_letters + s.digits + extra
    buf = ""
    while num:
        num, i = divmod(num, len(table))
        buf = table[i] + buf
    return buf


def split_tags(tags):
    "Parse a string and return a list of tags."
    return [t for t in tags.replace("\u3000", " ").split(" ") if t]


class Note:  # pylint: disable=R0902
    def __init__(self, col, model=None, note_id=None):
        assert not (model and note_id)
        self.col = col
        self.mod = None
        self.usn = None
        self.newlyAdded = None

        if note_id:
            self.id = note_id
            self.load()
        elif model:
            self.id = col.timestamp_for_table("notes")
            self.guid = guid64()
            self._model = model
            self.mid = model["id"]
            self.tags = []
            self.fields = [""] * len(self._model["flds"])
            self.flags = 0
            self.data = ""
            self._fmap = self.col.models.field_map(self._model)
            self.scm = self.col.scm

    def load(self):
        (self.guid, self.mid, self.mod, self.usn, self.tags, self.fields, self.flags, self.data) = self.col.db.first(
            f"""select guid, mid, modified, usn, tags, flds, flags, data
                from {self.col.username}.notes where id = %s""",
            self.id,
        )
        self.fields = splitFields(self.fields)
        self.tags = split_tags(self.tags)
        self._model = self.col.models.get(self.mid)
        self._fmap = self.col.models.field_map(self._model)
        self.scm = self.col.scm

    def load_from_db_row(self, row):
        (self.id, self.guid, self.mid, self.mod, self.usn, self.tags, self.fields, self.flags, self.data) = row
        self.fields = splitFields(self.fields)
        self.tags = split_tags(self.tags)
        self._model = self.col.models.get(self.mid)
        self._fmap = self.col.models.field_map(self._model)
        self.scm = self.col.scm
        return self

    def as_dict(self):
        return {"id": self.id, "model": self._model["name"], "fields": self.fields, "tags": self.tags}

    def __str__(self):
        return json.dumps(self.as_dict())

    def flush(self, mod=None):
        "If fields or tags have changed, write changes to disk."
        assert self.scm == self.col.scm
        self.newlyAdded = not self.col.db.scalar(f"select 1 from {self.col.username}.cards where nid = %s", self.id)

        sfld = stripHTMLMedia(self.fields[self._model["sortf"]])
        tags = self.col.tagstring_for_note(self.tags)

        fields = joinFields(self.fields)
        if not mod and self.col.db.scalar(
            f"select 1 from {self.col.username}.notes where id = %s and tags = %s and flds = %s", self.id, tags, fields
        ):
            # there has been no change, so nothing to do
            return

        csum = fieldChecksum(self.fields[0])
        self.mod = mod if mod else intTime()
        self.usn = self.col.usn
        sql = djankiserv.unki.AnkiDataModel.insert_on_conflict_update(self.col.username, "notes")
        self.col.db.execute(
            sql, self.id, self.guid, self.mid, self.mod, self.usn, tags, fields, sfld, csum, self.flags, self.data,
        )

        self.col.register_tags(self.tags)
        if not self.newlyAdded:
            self.col.gen_cards([self.id])

            # FIXME: see orig _postFlush method - should we do this here???
            # self.col.remEmptyCards(ids)

    def model(self):
        return self._model

    def items(self):
        return [(field["name"], self.fields[ordi]) for ordi, field in sorted(self._fmap.values())]

    def cards(self):
        card_ids = self.col.db.list(f"select id from {self.col.username}.cards where nid = %s order by ord", self.id)
        return [djankiserv.unki.cards.Card(self.col, card_id) for card_id in card_ids]

    def __getitem__(self, key):
        return self.fields[self._fmap[key][0]]

    def __setitem__(self, key, value):
        self.fields[self._fmap[key][0]] = value



# -*- coding: utf-8 -*-

import json
import logging
import os
import pathlib
import pickle
import subprocess

from django.conf import settings
from django.db import connection, connections
from django.http import HttpResponse

import djankiserv.unki
from djankiserv.assets import jsonfiles  # noqa: 401  # pylint: disable=W0611

from . import AnkiDataModelBase, get_data

MODEL_STD = 0
NEW_CARDS_DUE = 1

logger = logging.getLogger("djankiserv.unki.database")


def db_conn():
    if "userdata" in connections:
        return connections["userdata"]

    return connection


class StandardDB:
    def __init__(self):  # pylint: disable=W0231
        self._db = db_conn()
        self.mod = False

    @staticmethod
    def schema_exists(schema_name):
        with db_conn().cursor() as cur:
            cur.execute("SELECT 1 FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = %s", (schema_name,))
            res = cur.fetchone()
        return res[0] if res else 0

    @staticmethod
    def create_schema(schema_name):
        assert schema_name.isidentifier()
        with db_conn().cursor() as cur:
            for sql in djankiserv.unki.AnkiDataModel.generate_schema_sql_list(schema_name):
                if not sql.strip():
                    continue
                cur.execute(sql)
            res = cur.fetchone()
            return res[0]  # returns an Ok message
        # db_conn().commit()

    @staticmethod
    def delete_schema(schema_name):
        with db_conn().cursor() as cur:
            cur.execute(djankiserv.unki.AnkiDataModel.DROP_SCHEMA.replace("{schema_name}", schema_name))

    def execute(self, sql, *a, **ka):
        s = sql.strip().lower()
        # mark modified?
        for stmt in "insert", "update", "delete":
            if s.startswith(stmt):
                self.mod = True
        cur = self._db.cursor()
        if ka:
            cur.execute(sql, ka)
        else:
            cur.execute(sql, a)
        return cur

    def executemany(self, sql, li):
        self.mod = True
        cur = self._db.cursor()
        cur.executemany(sql, list(li))  # list() due to https://github.com/korfuri/django-prometheus/issues/240
        return cur

    def commit(self):
        pass
        # self._db.commit()

    def scalar(self, *a):
        res = self.execute(*a).fetchone()  # pylint: disable=E1120  # FIXME: how should I get rid of this disable?
        if res:
            return res[0]
        return None

    def first(self, *a):
        c = self.execute(*a)
        res = c.fetchone()
        c.close()
        return res

    def list(self, *a):
        return [x[0] for x in self.execute(*a)]  # pylint: disable=E1120  # FIXME: how should I get rid of this disable?

    def close(self):
        # FIXME: is this Ok?
        pass
        # self._db.commit()
        # self._db.close()


# for getting dumps of the DB to compare for unittests
def dump_io_to_file(session, method, io_obj, is_media=False):  # noqa: C901  # pylint: disable=R0914
    if not settings.DJANKISERV_GENERATE_TEST_ASSETS:  # we need this only for getting new test files
        return
    schema_name = session["name"]
    fname_dir = session["dump_base"] if not is_media else session["dump_base_media"]
    pathlib.Path(fname_dir).mkdir(parents=True, exist_ok=True)

    if isinstance(io_obj, HttpResponse):
        fname_base = os.path.join(fname_dir, "post_" + method)
        to_print = f"Response Content:\n{io_obj.content}\n"
        try:
            io_obj_json = json.loads(io_obj.content)
        except (json.decoder.JSONDecodeError, UnicodeDecodeError):
            io_obj_json = {"to_return": io_obj.content}  # this will get pickled below
    else:
        fname_base = os.path.join(fname_dir, "pre_" + method)
        to_print = f"POST:\n{io_obj.POST}\n\n"
        to_print += f"DATA:\n{get_data(io_obj)}\n"
        io_obj_json = get_data(io_obj)

    if os.path.exists(fname_base + ".txt"):
        i = 1
        while os.path.exists(f"{fname_base}.{i}.txt"):
            i += 1
        fname_base = f"{fname_base}.{i}"

    with open(fname_base + ".txt", "w") as fh:
        fh.write(to_print)

    proc_res = subprocess.run(
        ["pg_dump", db_conn().settings_dict["NAME"], "--schema", schema_name, "--column-inserts"],
        capture_output=True,
        text=True,
        check=True,
    )

    with open(fname_base + ".sql", "w") as fh:
        fh.write(proc_res.stdout)

    try:
        with open(fname_base + ".json", "w") as fh:
            json.dump(io_obj_json, fh, indent=4, sort_keys=True)
    except TypeError:
        # there is binary data in the json, so pickle instead
        with open(fname_base + ".bin", "wb") as fh:
            pickle.dump(io_obj_json, fh)
        try:  # to remove the broken .json file that was created
            os.remove(fname_base + ".json")
        except OSError:
            pass

    with open(fname_base + "_col.json", "w") as fh, db_conn().cursor() as cur:
        cur.execute(f"select conf, models, decks, dconf, tags from {schema_name}.col")
        row = cur.fetchall()
        if len(row) > 1:
            raise Exception("There should only ever be one collection, right?")
        conf = json.loads(row[0][0])
        models = json.loads(row[0][1])
        decks = json.loads(row[0][2])
        dconf = json.loads(row[0][3])
        tags = json.loads(row[0][4])
        fh.write(
            json.dumps(
                {"conf": conf, "models": models, "decks": decks, "dconf": dconf, "tags": tags},
                sort_keys=True,
                indent=4,
            )
        )


class PostgresAnkiDataModel(AnkiDataModelBase):
    MODEL_SETUP = ""

    MODEL_SETUP_LIST = MODEL_SETUP.split(";")
    CREATE_SCHEMA = "CREATE SCHEMA {schema_name};"
    DROP_SCHEMA = "DROP SCHEMA IF EXISTS {schema_name} CASCADE;"

    @staticmethod
    def generate_schema_sql_list(schema_name):
        sql = []
        sql += PostgresAnkiDataModel.MODEL_SETUP_LIST
        sql.append(PostgresAnkiDataModel.CREATE_SCHEMA.format(schema_name=schema_name))

        for table_name, defin in AnkiDataModelBase.MODEL.items():
            tsql = (
                f"CREATE TABLE {schema_name}.{table_name} ("
                + ",".join(
                    [f"{f['name']} {f['type']} {'' if f.get('nullable') else 'NOT'} NULL" for f in defin["fields"]]
                )
                + ");"
            )
            sql.append(tsql)
            identity = [x["name"] for x in defin["fields"] if "is_pk" in x and x["type"] == "bigint"]
            if identity:
                tsql = f"""
                ALTER TABLE {schema_name}.{table_name}
                    ALTER COLUMN {identity[0]} ADD GENERATED BY DEFAULT AS IDENTITY (
                        SEQUENCE NAME {schema_name}.{table_name}_{identity[0]}_seq
                        START WITH 1
                        INCREMENT BY 1
                        NO MINVALUE
                        NO MAXVALUE
                        CACHE 1
                    );"""

                sql.append(tsql)

            pk = [x["name"] for x in defin["fields"] if "is_pk" in x]
            if pk:
                tsql = f"""
                ALTER TABLE ONLY {schema_name}.{table_name}
                    ADD CONSTRAINT {table_name}_pkey PRIMARY KEY ({pk[0]});"""
                sql.append(tsql)

            for i in defin["indexes"]:
                tsql = (
                    f"CREATE INDEX {i['name']} ON {schema_name}.{table_name} ("  # defaults to btree for postgresql
                    + ",".join([x["name"] for x in i["fields"]])
                    + ");"
                )
                sql.append(tsql)

            if "initsql" in defin:
                tsql = defin["initsql"].replace("{schema_name}", schema_name)  # dynamic replace rather than fstring
                sql.append(tsql)

        sql.append(f"SELECT {AnkiDataModelBase.VERSION};")

        return sql

    @staticmethod
    def insert_on_conflict_update(schema_name, table_name):
        identity = [x["name"] for x in AnkiDataModelBase.MODEL[table_name]["fields"] if "is_pk" in x]
        fstr = ", ".join(["%s"] * len(AnkiDataModelBase.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in AnkiDataModelBase.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + f"ON CONFLICT ({identity[0]}) DO UPDATE SET "
            + ",".join(
                [
                    f"{f['name']} = EXCLUDED.{f['name']}"
                    for f in AnkiDataModelBase.MODEL[table_name]["fields"]
                    if "is_pk" not in f
                ]
            )
        )

    @staticmethod
    def insert_on_conflict_nothing(schema_name, table_name):
        identity = [x["name"] for x in AnkiDataModelBase.MODEL[table_name]["fields"] if "is_pk" in x]
        fstr = ", ".join(["%s"] * len(AnkiDataModelBase.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in AnkiDataModelBase.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + f"ON CONFLICT ({identity[0]}) DO NOTHING "
        )

    @staticmethod
    def replace_schema(cur, to_replace_name, replace_with_name):
        # rename the existing schema, rename the new schema to the username, delete the old schema
        cur.execute(f"ALTER SCHEMA {to_replace_name} RENAME TO {to_replace_name}_old")
        cur.execute(f"ALTER SCHEMA {replace_with_name} RENAME TO {to_replace_name}")
        cur.execute(f"DROP SCHEMA {to_replace_name}_old CASCADE")


class MariadbAnkiDataModel(AnkiDataModelBase):

    MODEL_SETUP = ""
    MODEL_SETUP_LIST = []

    CREATE_SCHEMA = "CREATE DATABASE {schema_name} CHARACTER SET utf8;"
    DROP_SCHEMA = "DROP DATABASE IF EXISTS {schema_name};"

    @staticmethod
    def generate_schema_sql_list(schema_name):
        sql = []
        sql += MariadbAnkiDataModel.MODEL_SETUP_LIST
        sql.append(MariadbAnkiDataModel.CREATE_SCHEMA.format(schema_name=schema_name))

        for table_name, defin in AnkiDataModelBase.MODEL.items():
            tsql = (
                f"CREATE TABLE {schema_name}.{table_name} ("
                + ",".join(
                    [f"{f['name']} {f['type']} {'' if f.get('nullable') else 'NOT'} NULL" for f in defin["fields"]]
                )
                + ");"
            )
            sql.append(tsql)
            identity = [x["name"] for x in defin["fields"] if "is_pk" in x and x["type"] == "bigint"]
            if identity:
                tsql = (
                    f"""ALTER TABLE {schema_name}.{table_name} MODIFY """
                    f"""{identity[0]} BIGINT NOT NULL AUTO_INCREMENT PRIMARY KEY"""
                )

                sql.append(tsql)
            else:
                pk = [x for x in defin["fields"] if "is_pk" in x]
                if pk:
                    pk_colspec = pk[0]["name"] + "(255)" if pk[0]["type"] == "text" else ""
                    tsql = f"""
                    ALTER TABLE {schema_name}.{table_name}
                        ADD CONSTRAINT {table_name}_pkey PRIMARY KEY ({pk_colspec});"""
                    sql.append(tsql)

            for i in defin["indexes"]:
                tsql = (
                    f"CREATE INDEX {i['name']} ON {schema_name}.{table_name} ("  # defaults to btree for postgresql
                    + ",".join([x["name"] + ("(255)" if x["type"] == "text" else "") for x in i["fields"]])
                    + ");"
                )
                sql.append(tsql)

            if "initsql" in defin:
                tsql = defin["initsql"].replace("{schema_name}", schema_name)  # dynamic replace rather than fstring
                sql.append(tsql)

            # sql += tsql
        sql.append(f"SELECT {AnkiDataModelBase.VERSION};")

        return sql

    @staticmethod
    def insert_on_conflict_update(schema_name, table_name):

        fstr = ", ".join(["%s"] * len(MariadbAnkiDataModel.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in MariadbAnkiDataModel.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + "ON DUPLICATE KEY UPDATE "
            + ",".join(
                [
                    f"{f['name']} = VALUES({f['name']})"
                    for f in MariadbAnkiDataModel.MODEL[table_name]["fields"]
                    if "is_pk" not in f
                ]
            )
        )

    @staticmethod
    def insert_on_conflict_nothing(schema_name, table_name):

        identity = [x["name"] for x in MariadbAnkiDataModel.MODEL[table_name]["fields"] if "is_pk" in x]
        fstr = ", ".join(["%s"] * len(MariadbAnkiDataModel.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in MariadbAnkiDataModel.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + f"ON DUPLICATE KEY UPDATE {identity[0]} = {identity[0]}"
        )

    @staticmethod
    def replace_schema(cur, to_replace_name, replace_with_name):
        # rename the existing schema, rename the new schema to the username, delete the old schema

        sql = f"""
            SELECT CONCAT('RENAME TABLE ',
                table_schema, '.', table_name, ' TO ', '{to_replace_name}.', table_name, ';') as commands
            FROM information_schema.TABLES
            WHERE table_schema = '{replace_with_name}';
            """
        cur.execute(sql)
        copy_commands = cur.fetchall()

        sql = f"DROP DATABASE {to_replace_name}; CREATE DATABASE {to_replace_name} CHARACTER SET utf8;"
        cur.execute(sql)

        for com in copy_commands:
            cur.execute(com[0])

        sql = f"DROP DATABASE {replace_with_name};"
        cur.execute(sql)



# -*- coding: utf-8 -*-

from django.conf import settings
from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
from rest_framework.decorators import api_view, parser_classes
from rest_framework.parsers import JSONParser

from djankiserv.unki.collection import Collection


@csrf_exempt
@api_view(["POST"])
@parser_classes([JSONParser])
def add_notes(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        note_ids = []
        for note in request.data["notes"]:
            note_ids.append(col.create_note(note, deck_name=request.data["deck"]))

    return JsonResponse({"note_ids": note_ids})


@csrf_exempt
@api_view(["POST", "GET"])
@parser_classes([JSONParser])
def notes(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        notes = col.get_notes(request.data.get("ids"))

    return JsonResponse({"notes": [n.as_dict() for n in notes]})


@csrf_exempt
@api_view(["POST"])
@parser_classes([JSONParser])
def delete_notes(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        note_ids = request.data.get("ids")
        if not note_ids:  # is this a little dangerous maybe?
            note_ids = col.all_note_ids()
        col.rem_notes(note_ids)

    return JsonResponse({"status": "ok"})


@csrf_exempt
@api_view(["POST", "GET"])
def decks(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        return JsonResponse({"decks": col.decks.decks})


@csrf_exempt
@api_view(["POST", "GET"])
def decks_conf(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        return JsonResponse({"decks_conf": col.decks.dconf})


@csrf_exempt
@api_view(["POST", "GET"])
def tags(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        return JsonResponse({"tags": list(col.tags.keys())})


@csrf_exempt
@api_view(["POST", "GET"])
def models(request):
    with Collection(request.user.username, settings.DJANKISERV_DATA_ROOT) as col:
        return JsonResponse({"models": col.models.models})



# -*- coding: utf-8 -*-
# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import copy
import json
import operator
import pkgutil
import unicodedata

from djankiserv.unki import ids2str, intTime  # intTime is never actually used, so all occurrences are dead code

REM_DECK = 2


class DeckManager:  # pylint: disable=R0904
    def __init__(self, col):
        self.col = col
        self.decks = None  # to make pylint happy
        self.dconf = None  # to make pylint happy
        self.changed = False  # to make pylint happy

    def load(self, decks, dconf):
        self.decks = json.loads(decks)
        self.dconf = json.loads(dconf)
        # set limits to within bounds
        found = False
        for c in list(self.dconf.values()):
            for t in ("rev", "new"):
                pd = "perDay"
                if c[t][pd] > 999999:
                    c[t][pd] = 999999
                    self.save(c)
                    found = True
        if not found:
            self.changed = False

    def save(self, g=None):
        "Can be called with either a deck or a deck configuration."
        if g:
            g["mod"] = intTime()
            g["usn"] = self.col.usn
        self.changed = True

    def flush(self):
        if self.changed:
            self.col.db.execute(
                f"update {self.col.username}.col set decks=%s, dconf=%s", json.dumps(self.decks), json.dumps(self.dconf)
            )
            self.changed = False

    def all(self):
        "A list of all decks."
        return list(self.decks.values())

    def get(self, did, default=True):
        sid = str(did)
        if sid in self.decks:
            return self.decks[sid]
        if default:
            return self.decks["1"]
        return None

    def update(self, g):
        "Add or update an existing deck. Used for syncing and merging."
        self.decks[str(g["id"])] = g
        self.maybe_add_to_active()
        # mark registry changed, but don't bump mod time
        self.save()

    def update_conf(self, g):
        self.dconf[str(g["id"])] = g
        self.save()

    def all_conf(self):
        "A list of all deck config."
        return list(self.dconf.values())

    def conf_for_did(self, did):
        deck = self.get(did, default=False)
        assert deck
        if "conf" in deck:
            conf = self.get_conf(deck["conf"])
            conf["dyn"] = False
            return conf
        # dynamic decks have embedded conf
        return deck

    def get_conf(self, confId):
        return self.dconf[str(confId)]

    def maybe_add_to_active(self):
        # reselect current deck, or default if current has disappeared
        c = self.current()
        self.select(c["id"])

    def _recover_orphans(self):
        dids = list(self.decks.keys())
        mod = self.col.db.mod
        self.col.db.execute(f"update {self.col.username}.cards set did = 1 where did not in " + ids2str(dids))
        self.col.db.mod = mod

    def _check_deck_tree(self):
        decks = self.col.decks.all()
        decks.sort(key=operator.itemgetter("name"))
        names = set()

        for deck in decks:
            # two decks with the same name?
            if deck["name"] in names:
                print("fix duplicate deck name", deck["name"].encode("utf8"))
                deck["name"] += "%d" % intTime(1000)
                self.save(deck)

            # ensure no sections are blank
            if not all(deck["name"].split("::")):
                print("fix deck with missing sections", deck["name"].encode("utf8"))
                deck["name"] = "recovered%d" % intTime(1000)
                self.save(deck)

            # immediate parent must exist
            if "::" in deck["name"]:
                immediateParent = "::".join(deck["name"].split("::")[:-1])
                if immediateParent not in names:
                    print("fix deck with missing parent", deck["name"].encode("utf8"))
                    self._ensure_parents(deck["name"])
                    names.add(immediateParent)

            names.add(deck["name"])

    def check_integrity(self):
        self._recover_orphans()
        self._check_deck_tree()

    # FIXME: this should be removed - it should have no relevance server-side!!!
    def active(self):
        "The currrently active dids. Make sure to copy before modifying."
        return self.col.conf["activeDecks"]

    # FIXME: this should be removed - it should have no relevance server-side!!!
    def selected(self):
        "The currently selected did."
        return self.col.conf["curDeck"]

    # FIXME: this should be removed - it should have no relevance server-side!!!
    def current(self):
        return self.get(self.selected())

    # FIXME: this should be removed - it should have no relevance server-side!!!
    # or should it???
    def select(self, did):
        "Select a new branch."
        # make sure arg is an int
        did = int(did)
        # current deck
        self.col.conf["curDeck"] = did
        # and active decks (current + all children)
        actv = self.children(did)
        actv.sort()
        self.col.conf["activeDecks"] = [did] + [a[1] for a in actv]
        self.changed = True

    def children(self, did):
        "All children of did, as (name, id)."
        name = self.get(did)["name"]
        actv = []
        for g in self.all():
            if g["name"].startswith(name + "::"):
                actv.append((g["name"], g["id"]))
        return actv

    def parents(self, did, deck_name_map=None):
        "All parents of did."
        # get parent and grandparent names
        parents = []
        for part in self.get(did)["name"].split("::")[:-1]:
            if not parents:
                parents.append(part)
            else:
                parents.append(parents[-1] + "::" + part)
        # convert to objects
        for c, p in enumerate(parents):
            if deck_name_map:
                deck = deck_name_map[p]
            else:
                deck = self.get(self.id(p))
            parents[c] = deck
        return parents

    def name_map(self):
        return dict((d["name"], d) for d in self.decks.values())

    def rem(self, did, cards_too=False, children_too=True):  # noqa: C901 # pylint: disable=R0912
        "Remove the deck. If cardsToo, delete any cards inside."
        if str(did) == "1":
            # we won't allow the default deck to be deleted, but if it's a
            # child of an existing deck then it needs to be renamed
            deck = self.get(did)
            if "::" in deck["name"]:
                base = deck["name"].split("::")[-1]
                suffix = ""
                while True:
                    # find an unused name
                    name = base + suffix
                    if not self.by_name(name):
                        deck["name"] = name
                        self.save(deck)
                        break
                    suffix += "1"
            return
        # log the removal regardless of whether we have the deck or not
        self.col.add_graves([did], REM_DECK)
        # do nothing else if doesn't exist
        if not str(did) in self.decks:
            return
        deck = self.get(did)
        if deck["dyn"]:
            # deleting a cramming deck returns cards to their previous deck
            # rather than deleting the cards
            self.col.sched.emptyDyn(did)
            if children_too:
                for name, ch_id in self.children(did):
                    self.rem(ch_id, cards_too)
        else:
            # delete children first
            if children_too:
                # we don't want to delete children when syncing
                for name, ch_id in self.children(did):
                    self.rem(ch_id, cards_too)
            # delete cards too?
            if cards_too:
                # don't use cids(), as we want cards in cram decks too
                cids = self.col.db.list(f"select id from {self.col.username}.cards where did=%s or odid=%s", did, did)
                self.col.rem_cards(cids)
        # delete the deck and add a grave
        del self.decks[str(did)]
        # ensure we have an active deck
        if did in self.active():
            self.select(int(list(self.decks.keys())[0]))
        self.save()

    def by_name(self, name):
        "Get deck with NAME."
        for m in list(self.decks.values()):
            if m["name"] == name:
                return m
        return None

    def get_or_add(self, name, create=True, dtype=None):
        "Add a deck with NAME. Reuse deck if already exists. Return id as int."
        if dtype is None:
            dtype = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_deck.json").decode("utf-8"))

        name = name.replace('"', "")
        name = unicodedata.normalize("NFC", name)
        for did, g in list(self.decks.items()):
            if unicodedata.normalize("NFC", g["name"].lower()) == name.lower():
                return int(did)
        if not create:
            return None
        g = copy.deepcopy(dtype)
        if "::" in name:
            # not top level; ensure all parents exist
            name = self._ensure_parents(name)
        g["name"] = name
        while 1:
            did = intTime(1000)
            if str(did) not in self.decks:
                break
        g["id"] = did
        self.decks[str(did)] = g
        self.save(g)
        self.maybe_add_to_active()
        return int(did)

    @staticmethod
    def _path(name):
        return name.split("::")

    def _ensure_parents(self, name):
        "Ensure parents exist, and return name with case matching parents."
        s = ""
        path = self._path(name)
        if len(path) < 2:
            return name
        for p in path[:-1]:
            if not s:
                s += p
            else:
                s += "::" + p
            # fetch or create
            did = self.get_or_add(s)
            # get original case
            s = self.name(did)
        name = s + "::" + path[-1]
        return name

    def name(self, did, default=False):
        deck = self.get(did, default=default)
        if deck:
            return deck["name"]
        return "[no deck]"

    # FIXME: maybe not called
    # def is_dyn(self, deck_id):
    #     return self.get(deck_id)['dyn']



# -*- coding: utf-8 -*-
# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

# flake8: noqa  # this will eventually be rewritten but for the moment just keep upstream
# pylint: disable=W,R

import time
from heapq import *
from operator import itemgetter

from djankiserv.unki import ids2str, intTime

# queue types: 0=new/cram, 1=lrn, 2=rev, 3=day lrn, -1=suspended, -2=buried
# revlog types: 0=lrn, 1=rev, 2=relrn, 3=cram
# positive revlog intervals are in days (rev), negative in seconds (lrn)

# copied from anki.consts
NEW_CARDS_DISTRIBUTE = 0


class Scheduler:
    name = "std"
    haveCustomStudy = True
    _spreadRev = True
    _burySiblingsOnAnswer = True

    def __init__(self, col):
        self.col = col
        self.queueLimit = 50
        self.reportLimit = 1000
        self.reps = 0
        self.today = None
        self._haveQueues = False
        self._updateCutoff()

    def reset(self):
        self._updateCutoff()
        self._resetLrn()
        self._resetRev()
        self._resetNew()
        self._haveQueues = True

    def counts(self, card=None):
        counts = [self.newCount, self.lrnCount, self.revCount]
        if card:
            idx = self.countIdx(card)
            if idx == 1:
                counts[1] += card.left // 1000
            else:
                counts[idx] += 1
        return tuple(counts)

    def unburyCards(self):
        "Unbury cards."
        self.col.conf["lastUnburied"] = self.today
        self.col.db.execute(f"update {self.col.username}.cards set queue=type where queue = -2")

    def _walkingCount(self, limFn=None, cntFn=None):
        tot = 0
        pcounts = {}
        # for each of the active decks
        name_map = self.col.decks.name_map()
        for did in self.col.decks.active():
            # early alphas were setting the active ids as a str
            did = int(did)
            # get the individual deck's limit
            lim = limFn(self.col.decks.get(did))
            if not lim:
                continue
            # check the parents
            parents = self.col.decks.parents(did, name_map)
            for p in parents:
                # add if missing
                if p["id"] not in pcounts:
                    pcounts[p["id"]] = limFn(p)
                # take minimum of child and parent
                lim = min(pcounts[p["id"]], lim)
            # see how many cards we actually have
            cnt = cntFn(did, lim)
            # if non-zero, decrement from parent counts
            for p in parents:
                pcounts[p["id"]] -= cnt
            # we may also be a parent
            pcounts[did] = lim - cnt
            # and add to running total
            tot += cnt
        return tot

    def deckDueList(self):
        "Returns [deckname, did, rev, lrn, new]"
        self._checkDay()
        self.col.decks.check_integrity()
        decks = self.col.decks.all()
        decks.sort(key=itemgetter("name"))
        lims = {}
        data = []

        def parent(name):
            parts = name.split("::")
            if len(parts) < 2:
                return None
            parts = parts[:-1]
            return "::".join(parts)

        for deck in decks:
            p = parent(deck["name"])
            # new
            nlim = self._deckNewLimitSingle(deck)
            if p:
                nlim = min(nlim, lims[p][0])
            new = self._newForDeck(deck["id"], nlim)
            # learning
            lrn = self._lrnForDeck(deck["id"])
            # reviews
            rlim = self._deckRevLimitSingle(deck)
            if p:
                rlim = min(rlim, lims[p][1])
            rev = self._revForDeck(deck["id"], rlim)
            # save to list
            data.append([deck["name"], deck["id"], rev, lrn, new])
            # add deck as a parent
            lims[deck["name"]] = [nlim, rlim]
        return data

    def _resetNewCount(self):
        cntFn = lambda did, lim: self.col.db.scalar(
            f"select least(count(0), %s) from {self.col.username}.cards where did = %s and queue = 0", lim, did,
        )
        self.newCount = self._walkingCount(self._deckNewLimitSingle, cntFn)

    def _resetNew(self):
        self._resetNewCount()
        self._newDids = self.col.decks.active()[:]
        self._newQueue = []
        self._updateNewCardRatio()

    def _updateNewCardRatio(self):
        if self.col.conf["newSpread"] == NEW_CARDS_DISTRIBUTE:
            if self.newCount:
                self.newCardModulus = (self.newCount + self.revCount) // self.newCount
                # if there are cards to review, ensure modulo >= 2
                if self.revCount:
                    self.newCardModulus = max(2, self.newCardModulus)
                return
        self.newCardModulus = 0

    def _newForDeck(self, did, lim):
        "New count for a single deck."
        if not lim:
            return 0
        lim = min(lim, self.reportLimit)
        return self.col.db.scalar(
            f"""select count(0) from {self.col.username}.cards where did = %s and queue = 0 limit %s""", did, lim
        )

    def _deckNewLimitSingle(self, g):
        "Limit for deck without parent limits."
        if g["dyn"]:
            return self.reportLimit
        c = self.col.decks.conf_for_did(g["id"])
        return max(0, c["new"]["perDay"] - g["newToday"][1])

    def _resetLrnCount(self):
        # sub-day
        self.lrnCount = int(
            self.col.db.scalar(
                f"""
                select sum(remaining/1000) from (select remaining from {self.col.username}.cards where
                did in {self._deckLimit()} and queue = 1 and due < %s limit {self.reportLimit}) as foo""",
                self.dayCutoff,
            )
            or 0
        )
        # day
        self.lrnCount += self.col.db.scalar(
            f"""
            select count(0) from {self.col.username}.cards where did in {self._deckLimit()} and queue = 3
            and due <= %s limit {self.reportLimit}""",
            self.today,
        )

    def _resetLrn(self):
        self._resetLrnCount()
        self._lrnQueue = []
        self._lrnDayQueue = []
        self._lrnDids = self.col.decks.active()[:]

    def _lrnForDeck(self, did):
        cnt = int(
            self.col.db.scalar(
                f"""select sum(remaining/1000) from
                        (select remaining from {self.col.username}.cards
                            where did = %s and queue = 1 and due < %s limit %s) as foo""",
                did,
                intTime() + self.col.conf["collapseTime"],
                self.reportLimit,
            )
            or 0
        )
        return cnt + self.col.db.scalar(
            f"""select count(0) from
                {self.col.username}.cards where did = %s and queue = 3
                and due <= %s limit %s""",
            did,
            self.today,
            self.reportLimit,
        )

    def _deckRevLimitSingle(self, d):
        if d["dyn"]:
            return self.reportLimit
        c = self.col.decks.conf_for_did(d["id"])
        return max(0, c["rev"]["perDay"] - d["revToday"][1])

    def _revForDeck(self, did, lim):
        lim = min(lim, self.reportLimit)
        return self.col.db.scalar(
            f""" select count(0) from {self.col.username}.cards where did = %s and queue = 2
                and due <= %s limit %s""",
            did,
            self.today,
            lim,
        )

    def _resetRevCount(self):
        def cntFn(did, lim):
            return self.col.db.scalar(
                f"""
                select count(0) from {self.col.username}.cards where
                did = %s and queue = 2 and due <= %s limit {lim}""",
                did,
                self.today,
            )

        self.revCount = self._walkingCount(self._deckRevLimitSingle, cntFn)

    def _resetRev(self):
        self._resetRevCount()
        self._revQueue = []
        self._revDids = self.col.decks.active()[:]

    def _deckLimit(self):
        return ids2str(self.col.decks.active())

    def _updateCutoff(self):
        oldToday = self.today
        # days since col created
        self.today = int((time.time() - self.col.crt) // 86400)
        # end of day cutoff
        self.dayCutoff = self.col.crt + (self.today + 1) * 86400
        # update all daily counts, but don't save decks to prevent needless
        # conflicts. we'll save on card answer instead
        def update(g):
            for t in "new", "rev", "lrn", "time":
                key = t + "Today"
                if g[key][0] != self.today:
                    g[key] = [self.today, 0]

        for deck in self.col.decks.all():
            update(deck)
        # unbury if the day has rolled over
        unburied = self.col.conf.get("lastUnburied", 0)
        if unburied < self.today:
            self.unburyCards()

    def _checkDay(self):
        # check if the day has rolled over
        if time.time() > self.dayCutoff:
            self.reset()

    def emptyDyn(self, deck_id, limit=None):
        if not limit:
            limit = f"did = {deck_id}"
        # move out of cram queue
        self.col.db.execute(
            f"""
            update cards set did = odid, queue = (case when type = 1 then 0
            else type end), type = (case when type = 1 then 0 else type end),
            due = odue, odue = 0, odid = 0, usn = %s where {limit}""",
            self.col.usn,
        )



# -*- coding: utf-8 -*-

import logging
import os
import tempfile
import time
from sqlite3 import dbapi2 as sqlite

import djankiserv.unki
from djankiserv.assets import jsonfiles  # noqa: 401
from djankiserv.unki import REM_CARD, REM_NOTE, ids2str, intTime
from djankiserv.unki.database import StandardDB, db_conn
from djankiserv.unki.download import DB, sqlite3_for_download

logger = logging.getLogger("djankiserv.sync")


class SyncCollectionHandler:  # pylint: disable=R0904
    operations = ["meta", "applyChanges", "start", "applyGraves", "chunk", "applyChunk", "sanityCheck2", "finish"]

    def __init__(self, col, session=None):
        self.col = col
        if session:
            ## make sure not to create the property if it isn't in the dict
            ## this way we know the actual flow of events as it will raise an
            ## error. I know, this is nastiness...
            if "min_usn" in session:
                self.min_usn = session["min_usn"]
            if "max_usn" in session:
                self.max_usn = session["max_usn"]
            if "lnewer" in session:
                self.lnewer = session["lnewer"]

    ##
    ## Public API methods
    ##
    # ankidesktop >=2.1rc2 sends graves in applyGraves, but still expects
    # server-side deletions to be returned by start
    def start(self, min_usn, lnewer, offset):
        # FIXME: the clients now send an offset for the V1 scheduler, so we need to find out how
        # to distinguish between the V1 and V2 schedulers and only allow V1 for the moment
        if False and offset is not None:
            raise NotImplementedError(
                "You are using the experimental V2 scheduler, which is not supported by the server."
            )
        self.max_usn = self.col.usn
        self.min_usn = min_usn
        self.lnewer = not lnewer
        lgraves = self.removed()
        return lgraves

    def applyGraves(self, chunk):
        self.remove(chunk)

    def applyChanges(self, changes):
        lchg = self.changes()  # also has update side-effects, so must do before merge_changes
        # merge our side before returning
        self.merge_changes(changes)
        return lchg

    def applyChunk(self, chunk):
        if "revlog" in chunk:
            self.merge_revlog(chunk["revlog"])
        if "cards" in chunk:
            self.merge_cards(chunk["cards"])
        if "notes" in chunk:
            self.merge_notes(chunk["notes"])

    def all_data_to_sync_down(self):
        buf = dict(done=True)
        for table in ["revlog", "cards", "notes"]:
            cursor = self.cursor_for_table(table)
            rows = cursor.fetchall()
            # mark the objects as having been sent
            self.col.db.execute(f"update {self.col.username}.{table} set usn=%s where usn=-1", self.max_usn)
            buf[table] = rows
        return buf

    def sanityCheck2(self, client):
        server = self.sanity_check()
        if client != server:
            return dict(status="bad", c=client, s=server)
        return dict(status="ok")

    def finish(self):
        mod = intTime(1000)
        self.col.ls = mod
        self.col.usn = self.max_usn + 1
        # ensure we save the mod time even if no changes made
        self.col.db.mod = True
        self.col.save(mod=mod)
        return mod

    ##
    ## Project-internal methods
    ##
    def changes(self):
        "Bundle up small objects."
        d = dict(models=self.get_models(), decks=self.get_decks(), tags=self.get_tags())
        if self.lnewer:
            d["conf"] = self.get_conf()
            d["crt"] = self.col.crt
        return d

    def merge_changes(self, rchg):
        # then the other objects
        self.merge_models(rchg["models"])
        self.merge_decks(rchg["decks"])
        self.merge_tags(rchg["tags"])
        if "conf" in rchg:
            self.merge_conf(rchg["conf"])
        # this was left out of earlier betas
        if "crt" in rchg:
            self.col.crt = rchg["crt"]

    def sanity_check(self):  # noqa: C901
        if not self.col.basic_check():
            return "failed basic check"
        for t in "cards", "notes", "revlog", "graves":
            if self.col.db.scalar(f"select count(0) from {self.col.username}.{t} where usn = -1"):
                return "%s had usn = -1" % t
        for g in self.col.decks.all():
            if g["usn"] == -1:
                return "deck had usn = -1"
        for t, usn in self.col.all_tags():
            if usn == -1:
                return "tag had usn = -1"
        found = False
        for m in self.col.models.all():
            if m["usn"] == -1:
                return "model had usn = -1"
        if found:
            self.col.models.needs_saving = True

        self.col.sched.reset()
        # check for missing parent decks
        self.col.sched.deckDueList()
        # return summary of deck
        return [
            list(self.col.sched.counts()),
            self.col.db.scalar(f"select count(0) from {self.col.username}.cards"),
            self.col.db.scalar(f"select count(0) from {self.col.username}.notes"),
            self.col.db.scalar(f"select count(0) from {self.col.username}.revlog"),
            self.col.db.scalar(f"select count(0) from {self.col.username}.graves"),
            len(self.col.models.all()),
            len(self.col.decks.all()),
            len(self.col.decks.all_conf()),
        ]

    def cursor_for_table(self, table):
        if table == "revlog":
            return self.col.db.execute(
                f"""
                select id, cid, {self.max_usn}, ease, ivl, lastIvl, factor, rtime, type
                from {self.col.username}.revlog where usn >= {self.min_usn}"""
            )
        if table == "cards":
            return self.col.db.execute(
                f"""
                select id, nid, did, ord, modified, {self.max_usn}, type, queue, due, ivl, factor, reps,
                lapses, remaining, odue, odid, flags, data from {self.col.username}.cards where usn >= {self.min_usn}"""
            )
        return self.col.db.execute(
            f"""
            select id, guid, mid, modified, {self.max_usn}, tags, flds, '', '', flags, data
            from {self.col.username}.notes where usn >= {self.min_usn}"""
        )

    def remove(self, graves):
        # notes first, so we don't end up with duplicate graves
        self.col.rem_notes(graves["notes"])
        # then cards
        self.col.rem_cards(graves["cards"], notes=False)
        # and decks
        for oid in graves["decks"]:
            self.col.decks.rem(oid, children_too=False)

    def merge_models(self, rchg):
        for r in rchg:
            left = self.col.models.get(r["id"])
            # if missing locally or server is newer, update
            if not left or r["mod"] > left["mod"]:
                self.col.models.update(r)

    def merge_decks(self, rchg):
        for r in rchg[0]:
            left = self.col.decks.get(r["id"], False)
            # work around mod time being stored as string
            if left and not isinstance(left["mod"], int):
                left["mod"] = int(left["mod"])

            # if missing locally or server is newer, update
            if not left or r["mod"] > left["mod"]:
                self.col.decks.update(r)
        for r in rchg[1]:
            try:
                left = self.col.decks.get_conf(r["id"])
            except KeyError:
                left = None
            # if missing locally or server is newer, update
            if not left or r["mod"] > left["mod"]:
                self.col.decks.update_conf(r)

    def merge_tags(self, tags):
        self.col.register_tags(tags, usn=self.max_usn)

    def merge_revlog(self, logs):
        # import AnkiDataModel  # noqa # pylint: disable=C0415,W0611
        sql = djankiserv.unki.AnkiDataModel.insert_on_conflict_nothing(self.col.username, "revlog")

        self.col.db.executemany(sql, logs)

    def newer_rows(self, data, table, modIdx):
        ids = (r[0] for r in data)
        lmods = {}
        ids_str = ids2str(ids)
        for tid, mod in self.col.db.execute(
            f"select id, modified from {self.col.username}.{table} where id in {ids_str} and usn >= {self.min_usn}"
        ):
            lmods[tid] = mod
        update = []
        for r in data:
            if r[0] not in lmods or lmods[r[0]] < r[modIdx]:
                update.append(r)
        return update

    def merge_cards(self, cards):
        sql = djankiserv.unki.AnkiDataModel.insert_on_conflict_update(self.col.username, "cards")
        rows = self.newer_rows(cards, "cards", 4)
        self.col.db.executemany(sql, rows)

    def merge_notes(self, notes):
        rows = self.newer_rows(notes, "notes", 3)
        sql = djankiserv.unki.AnkiDataModel.insert_on_conflict_update(self.col.username, "notes")
        self.col.db.executemany(sql, rows)
        self.col.update_field_cache([f[0] for f in rows])

    def get_conf(self):
        return self.col.conf

    def merge_conf(self, conf):
        if conf != self.col.conf:
            self.col.conf = conf
            self.col.db.mod = True

    # @staticmethod
    # def _old_client(cv):
    #     if not cv:
    #         return False

    #     note = {"alpha": 0, "beta": 0, "rc": 0}
    #     client, version, platform = cv.split(',')

    #     for name in note.keys():
    #         if name in version:
    #             vs = version.split(name)
    #             version = vs[0]
    #             note[name] = int(vs[-1])

    #     # convert the version string, ignoring non-numeric suffixes like in beta versions of Anki
    #     version_nosuffix = re.sub(r'[^0-9.].*$', '', version)
    #     version_int = [int(x) for x in version_nosuffix.split('.')]

    #     if client == 'ankidesktop':
    #         return version_int < [2, 0, 27]
    #     elif client == 'ankidroid':
    #         if version_int == [2, 3]:
    #            if note["alpha"]:
    #               return note["alpha"] < 4
    #         else:
    #            return version_int < [2, 2, 3]
    #     else:  # unknown client, assume current version
    #         return False

    def removed(self):
        cards = []
        notes = []
        decks = []

        curs = self.col.db.execute(f"select oid, type from {self.col.username}.graves where usn >= %s", self.min_usn)

        for oid, t in curs:
            if t == REM_CARD:
                cards.append(oid)
            elif t == REM_NOTE:
                notes.append(oid)
            else:
                decks.append(oid)

        return dict(cards=cards, notes=notes, decks=decks)

    def get_models(self):
        return [m for m in self.col.models.all() if m["usn"] >= self.min_usn]

    def get_decks(self):
        return [
            [g for g in self.col.decks.all() if g["usn"] >= self.min_usn],
            [g for g in self.col.decks.all_conf() if g["usn"] >= self.min_usn],
        ]

    def get_tags(self):
        return [t for t, usn in self.col.all_tags() if usn >= self.min_usn]


def full_upload(data, username):
    # from sqlite to a standard db
    # TODO, maybe make this from config or an envvar
    insert_cursor_size = 10000  # this has no effect on memory and less than 1k significantly increases the time

    if not len(data) > 0:
        raise Exception("There is no data in the uploaded file")

    # write data to a tempfile
    with tempfile.NamedTemporaryFile(suffix=".anki2", delete=False) as f:
        temp_db_path = f.name
        f.write(data)

    # Verify integrity of the received database file before replacing our existing db.
    _check_sqlite3_db(temp_db_path)

    # create new schema in pg db that we'll fill with the new data
    timestamp = str(time.time()).replace(".", "_")
    tmp_schema_name = f"{username}_{timestamp}"
    StandardDB.create_schema(tmp_schema_name)

    # should maybe create a global transaction here?
    # same issue as https://github.com/tsudoko/anki-sync-server/issues/6
    with sqlite.connect(temp_db_path) as sqlite_conn, db_conn().cursor() as to_std_cursor:
        # clean the default values from the collection table
        to_std_cursor.execute(f"TRUNCATE TABLE {tmp_schema_name}.col")

        for table, props in djankiserv.unki.AnkiDataModel.MODEL.items():
            if props["parent"] != djankiserv.unki.AnkiDataModel.COLLECTION_PARENT_DB:
                continue

            start_sqlite_cursor = sqlite_conn.cursor()
            start_sqlite_cursor.execute(f"SELECT * FROM {table}")
            while True:
                current_data = start_sqlite_cursor.fetchmany(insert_cursor_size)
                if not current_data:
                    break
                # FIXME: psycopg2.extras is MUCH better than executemany, but is pgsql-specific
                # psycopg2.extras.execute_values(
                #     to_std_cursor, f"INSERT INTO {tmp_schema_name}.{table} VALUES %s", current_data)

                cols = ",".join(["%s"] * len(current_data[0]))
                to_std_cursor.executemany(f"INSERT INTO {tmp_schema_name}.{table} VALUES ({cols})", current_data)
            start_sqlite_cursor.close()
        djankiserv.unki.AnkiDataModel.replace_schema(to_std_cursor, username, tmp_schema_name)
        os.remove(temp_db_path)

    return "OK"


def full_download(col, username):
    select_cursor_size = 10000  # this has no effect on memory and less than 1k significantly increases the time

    download_db = _create_empty_sqlite3_db()

    # should maybe create a global transaction here?
    # same issue as https://github.com/tsudoko/anki-sync-server/issues/6
    col.save()

    with sqlite.connect(download_db) as sqlite_conn:
        sqlite_conn.isolation_level = "EXCLUSIVE"
        for table, props in djankiserv.unki.AnkiDataModel.MODEL.items():
            if props["parent"] != djankiserv.unki.AnkiDataModel.COLLECTION_PARENT_DB:
                continue

            with db_conn().cursor() as from_std_cursor:
                to_sqlite_cursor = sqlite_conn.cursor()  # can't use `with` with an sqlite3 cursor

                pgcols = ",".join([x["name"] for x in props["fields"]])

                from_std_cursor.execute(f"SELECT {pgcols} FROM {username}.{table}")
                while True:
                    current_data = from_std_cursor.fetchmany(select_cursor_size)

                    if not current_data:
                        break
                    cols = ",".join(["?"] * len(current_data[0]))
                    # TODO: executemany has terrible performance but whatever
                    to_sqlite_cursor.executemany(f"INSERT INTO {table} VALUES ({cols})", current_data)

                to_sqlite_cursor.close()

    _check_sqlite3_db(download_db)

    data = open(download_db, "rb").read()
    os.remove(download_db)

    return data


# using anki's own methods for empty db creation
def _create_empty_sqlite3_db():
    # we need to get a tmp filename where the file doesn't exist, or we can't use anki.storage.Collection
    # to create it
    tmp = tempfile.mktemp(suffix=".anki2")

    sqlite3_for_download(tmp)  # .close()  # will create and close cleanly

    # clean up the rows added in up-upstream db initialisation and leave only the schema
    #
    # conn.isolation_level = None and cur.execute("pragma journal_mode = delete")
    # are both required or the sqlite3 file will stay in WAL mode and not flush the transactions to disk
    with sqlite.connect(tmp) as conn:
        cur = conn.cursor()  # can't use `with` with an sqlite3 cursor
        cur.execute("DELETE FROM col")
        cur.close()

    return tmp


def _check_sqlite3_db(db_path):
    try:
        with DB(db_path) as test_db:
            if test_db.scalar("pragma integrity_check") != "ok":
                raise Exception("Integrity check failed for uploaded collection database file.")
    except sqlite.Error as err:
        raise Exception("Uploaded collection database file is corrupt.") from err



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['djankiserv',
 'djankiserv.api',
 'djankiserv.assets.jsonfiles',
 'djankiserv.assets.sql.sqlite3',
 'djankiserv.migrations',
 'djankiserv.sync',
 'djankiserv.unki']

package_data = \
{'': ['*']}

install_requires = \
['django-k8s>=0.2.9,<0.3.0',
 'django>=3.1,<4.0',
 'djangorestframework-simplejwt>=4.4.0,<5.0.0',
 'djangorestframework>=3.11.1,<4.0.0',
 'gunicorn>=20.0.4,<21.0.0']

extras_require = \
{'mysql': ['mysqlclient>=2.0.1,<3.0.0'],
 'pgsql': ['psycopg2-binary>=2.8.6,<3.0.0']}

setup_kwargs = {
    'name': 'djankiserv',
    'version': '0.0.9',
    'description': 'Django-based synchronisation and API server for Anki',
    'long_description': '# djankiserv\n\n`djankiserv` is an open source Django-based implementation of a synchronisation server for Anki 2.1+. It includes a user manager (the native Django user system).\n\n[Installation](doc/Installation.md) -\xa0[Connecting Anki to the sync server](doc/ConnectingAnki.md) - [Development](doc/Development.md) - [Contributing](doc/CONTRIBUTING.md)\n\n## About this implementation\n\nThis implementation was initially developed in order to support the spaced repetition functionality for [`Transcrobes`](https://transcrob.es), an open source language learning platform/ecosystem.\n\nAny requests or functionality that don\'t interfere with using this project for that will definitely be entertained. Ideally the server would do everything that Ankiweb does, and much more. PRs are obviously always welcome!\n\n### Technical differences\n\nUnlike the other popular open source Anki synchronisation server [`anki-sync-server`](https://github.com/ankicommunity/anki-sync-server), `djankiserv` stores the user data in a "proper" RDBMS. There are two \'database connections\' that can be set - those for the \'system\' (sessions, users, etc.) and those for user data. The \'system\' stuff is just plain old Django, so any supported database can be used. The user data part currently uses either `postgresql` schemas or `mysql` databases, and currently only supports those two, though supporting other RDBMSes will definitely be considered later. `sqlite3` is an embedded database and works great for that. It is not appropriate for use in modern web applications in the opinion of the maintainer, so will never be supported by `djankiserv`.\n\nThere is a basic API for getting certain, per-user collection-related information (decks, deck configuration, models, tags) and also `notes` for a given user. It may evolve to include other functions, statistics and even doing cards, though the focus is currently on getting and maintaining proper synchronisation as well as the basic API for `notes`.\n\n### Limitations\n\nThis is alpha software with some occasional data loss bugs. It works, sorta, if you hold it right. If it kills your kittens then you were forewarned!\n\nCurrent known limitations (bugs!):\n\n- it doesn\'t support abort and if it crashes in the middle of a sync then the server will have a corrupt view of the database. You should force an upload sync on next synchronisation if this ever happens!\n- The v2 scheduler is not supported, and it is unclear how difficult this might be to implement.\n',
    'author': 'Anton Melser',
    'author_email': 'anton@melser.org',
    'maintainer': 'Anton Melser',
    'maintainer_email': 'anton@melser.org',
    'url': 'https://github.com/ankicommunity/djankiserv',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)
